import logging
from pprint import pprint
from uuid import uuid4
import uuid
from orchestrator.exceptions import OrchestratorMissingParam
from orchestrator.orchestrator_http import OrchestratorHTTP
import requests
from urllib.parse import urlencode
from orchestrator.orchestrator_queue_item import QueueItem

__all__ = ["Queue"]


class Queue(OrchestratorHTTP):
    classes = ["Comments", "Status", "Reference", "SpecificContent"]
    """
    Constructor. 

    @client_id: the client id 
    @refresh_token: a refresh token  
    @tenant_name: account's logical name
    @folder_id: the folder id 
    @folder_name: the folder name
    @session: a session object (options)
    @queue_name: the queue name
    @queue_id: the queue id
    """

    def __init__(self, client_id, refresh_token, tenant_name, folder_id=None, folder_name=None, session=None, queue_name=None, queue_id=None, access_token=None):
        super().__init__(client_id=client_id, refresh_token=refresh_token, tenant_name=tenant_name, folder_id=folder_id, session=session)
        if not queue_id:
            raise OrchestratorMissingParam(value="queue_id",
                                           message="Required parameter(s) missing: queue_id")
        self.id = queue_id
        self.name = queue_name
        self.folder_name = folder_name
        self.folder_id = folder_id
        self.tenant_name = tenant_name
        self.base_url = f"{self.cloud_url}/{self.tenant_name}/JTBOT/odata"
        self.access_token = access_token
        self.refresh_token = refresh_token
        if session:
            self.session = session
        else:
            self.session = requests.Session()

    def __str__(self):
        return f"Queue Id: {self.id} \nQueue Name: {self.name} \nFolder Id: {self.folder_id} \nFolder Name: {self.folder_name}"

    def info(self):
        """
            Returns information about the queue

            @returns: dictionary with more in depth information
            about the queue 
        """
        endpoint = f"/QueueDefinitions({self.id})"
        url = f"{self.base_url}{endpoint}"
        data = self._get(url)
        return data

    def start(self, machine_identifier, specific_content=None, references=None, separator="-", fields=None):
        """
            Starts a given transaction

            @machine_identifier: the machine's unique identifier
            @specific_content: the specific content of the transaction
            @reference: a reference from the specific content 
            @fields: a dictionary of additional fields to be added to the specific content
        """
        ran_uuid = str(uuid4())
        batch_id = str(uuid.uuid4())
        logging.debug("Starting new transaction")
        endpoint = "/Queues/UiPathODataSvc.StartTransaction"
        format_body_start = {
            "transactionData": {
                "Name": self.name,
                "RobotIdentifier": machine_identifier,
                "SpecificContent": specific_content,
            }
        }
        if references:
            try:
                ref = ""
                for reference in references:
                    value = format_body_start["transactionData"]["SpecificContent"][reference]
                    ref += value+separator
                format_body_start["transactionData"]["Reference"] = f"{ref[:-1]}#{batch_id}"
            except KeyError as err:
                if reference in err.args:
                    logging.error(f"Invalid reference: {reference} not found in SpecificContent")
                    raise Exception(f"Invalid reference: {reference} not found in SpecificContent")
            format_body_start["transactionData"]["SpecificContent"]["ItemID"] = value
            format_body_start["transactionData"]["SpecificContent"]["ReferenceID"] = ran_uuid
            format_body_start["transactionData"]["SpecificContent"]["BatchID"] = batch_id
        if fields:
            format_body_start["transactionData"]["SpecificContent"].update(fields)

        url = f"{self.base_url}{endpoint}"
        return self._post(url, body=format_body_start)

    def get_processing_records(self, num_days=1, options=None):
        """
            Returns a list of processing records for a given
            queue and a certain number of days (by default, hourly reports
            from the last day)

            @num_days: the number of days before today from which to get
            the processing records (default: 1)
            @options: dictionary of odata filtering options
        """
        endpoint = "/QueueProcessingRecords"
        query = f"daysNo={num_days},queueDefinitionId={self.id}"
        uipath_svc = f"/UiPathODataSvc.RetrieveLastDaysProcessingRecords({query})"
        if options:
            query_params = urlencode(options)
            url = f"{self.base_url}{endpoint}?{query_params}"
        else:
            url = f"{self.base_url}{endpoint}{uipath_svc}"
        data = self._get(url)
        return data['value']

    def get_item_by_id(self, item_id):
        """
            Gets a single Item by item id

            @item_id: the id of the item
            ========
            @returns: an Item object with the specified item id
        """
        return QueueItem(self.client_id, self.refresh_token, self.tenant_name, self.folder_id, self.folder_name, self.name, self.id, self.session, item_id, access_token=self.access_token)

    def get_queue_items(self, options=None):
        """
            Returns a list of queue items of the given queue

            @options: dictionary of odata filtering options ($filter tag will be overwritten)
            =========
            @returns: a list of QueueItem objects of the given queue (Maximum number of results: 1000)
        """
        endpoint = "/QueueItems"
        if options and ("$filter" in options):
            print("true")
            odata_filter = {"$filter": f"QueueDefinitionId eq {self.id} and {options['$filter']}"}
        else:
            odata_filter = {"$Filter": f"QueueDefinitionId eq {self.id}"}
        # if options:
        #     odata_filter.update(options)
        query_params = urlencode(odata_filter)
        url = f"{self.base_url}{endpoint}?{query_params}"
        data = self._get(url)
        # pprint(data)
        filt_data = data['value']
        return [QueueItem(self.client_id, self.refresh_token, self.tenant_name, self.folder_id, self.folder_name, self.name, self.id, session=self.session, item_id=item["Id"], content=item["SpecificContent"], reference=item["Reference"], access_token=self.access_token) for item in filt_data]

    def get_queue_items_by_status(self, status):
        """
        Returns a list of QueueItems with the status 
        as indicated in the argument.
        """
        endpoint = "/QueueItems"
        odata_filter = {"$filter": f"QueueDefinitionId eq {self.id} and Status eq '{status}'"}
        query_params = urlencode(odata_filter)
        url = f"{self.base_url}{endpoint}?{query_params}"
        data = self._get(url)
        filt_data = data['value']
        return [QueueItem(self.client_id, self.refresh_token, self.tenant_name, self.folder_id, self.folder_name, self.name, self.id, session=self.session, item_id=item["Id"], content=item["SpecificContent"], reference=item["Reference"], access_token=self.access_token) for item in filt_data]

    def _get_sp_contents(self, options=None):
        items = self.get_queue_items(options=options)
        contents = []
        for item in items:
            contents.append(item.content)
        return contents

    def _get_references(self, options=None):
        items = self.get_queue_items(options=options)
        references = []
        for item in items:
            references.append(item.reference)
        return references

    def check_duplicate(self, reference):
        """
        Given a queue reference, it checks whether a given queue 
        has already appeared in the queue once.

        Parameters:

            - `param` reference: the reference or part of it of the new queue item

        Returns: 
            If a reference is found in the queue with the given status, it returns the first
            item whose reference matches the one indicated as an argument. Otherwise it returns
            False.
        """
        filt_items = self.get_queue_items(options={"$filter": f"contains(Reference, '{reference}') and Status eq 'Successful'"})

        if len(filt_items) > 0:
            return filt_items[0]
        else:
            return False

    def get_queue_items_ids(self, options=None):
        """
            Returns a list of dictionaries with the queue 
            item ids 

            @options: dictionary of odata filtering options 
            ========
            @returns: a dictionary where the keys are the queue item
            ids of the given queue and the values the queue name
        """
        items = self.get_queue_items(options)
        ids = {}
        for item in items:
            ids.update({item.id: item.queue_name})
        return ids

    def add_queue_item(self, specific_content=None, priority="Low"):
        """Creates a new Item

            @specific_content: dictionary of key value pairs (it does not
            admit nested dictionaries; for it to work json.dumps first)

            @priority - sets up the priority (Low by default)
        """
        endpoint = "/Queues"
        uipath_svc = "/UiPathODataSvc.AddQueueItem"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        if not specific_content:
            raise OrchestratorMissingParam(value="specific_content", message="specific content cannot be null")
        format_body_queue = {
            "itemData": {
                "Priority": priority,
                "Name": self.name,
                "SpecificContent": specific_content,
                "Reference": self.generate_reference(),
                # "Progress": "In Progres"
            }
        }
        # pprint(format_body_queue)
        return self._post(url, body=format_body_queue)

    def _format_specific_content(self, sp_content, reference=None, priority="Low", progress="New", batch_id=None):
        ran_uuid = str(uuid4())
        if reference:
            try:
                ref_uuid = {"Reference": f"{sp_content[reference]}#{batch_id}"}
                sp_content.update({"ReferenceID": ran_uuid})
                sp_content.update({"BatchID": batch_id})
                sp_content.update({"ItemID": sp_content[reference]})
                formatted_sp_content = {
                    "Name": self.name,
                    "Priority": priority,
                    "SpecificContent": sp_content,
                    "Progress": progress,
                }
                formatted_sp_content.update(ref_uuid)
                return formatted_sp_content
            except KeyError as err:
                if reference in err.args:
                    raise Exception(f"Invalid reference: {reference} not found in sp_content")
        else:
            ref_uuid = {"Reference": f"{ran_uuid}"}
            sp_content.update({"ReferenceID": ran_uuid})
            sp_content.update({"BatchID": batch_id})
            formatted_sp_content = {
                "Name": self.name,
                "Priority": priority,
                "SpecificContent": sp_content,
                "Progress": progress,
            }
            formatted_sp_content.update(ref_uuid)
            return formatted_sp_content

    def bulk_create_items(self, specific_contents=None, priority="Low", progress="New", reference=None):
        """Adds a list of items for a given queue
            @param specific_content: dictionary of key value pairs. It does not
            admit nested dictionaries. If you want to be able to pass a dictionary 
            as a key value pair inside the specific content attribute, you need to 
                                json.dumps(dict) 
            first for it to work.
            @priority: sets up the priority (default: Low)
            @progress: sets up the progress bar (default: New)
            @reference: indicates a specific field of the specific content to
            be used as a queue reference.

            Specific Content includes by default the following columns:
                - BatchID: representing a unique ID for the specific batch of items to be uploaded
                - ReferenceID: a unique ID of the item
                - ItemID: if reference is set to true, ads a new field with the reference
        """
        endpoint = "/Queues"
        uipath_svc = "/UiPathODataSvc.BulkAddQueueItems"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        if not specific_contents:
            raise OrchestratorMissingParam(value="specific_contents", message="specific contents cannot be null")

        batch_id = str(uuid4())
        format_body_queue = {
            "commitType": "StopOnFirstFailure",
            "queueName": self.name,
            "queueItems": [self._format_specific_content(queue_name=self.name, sp_content=sp_content, reference=reference, priority=priority, progress=progress, batch_id=batch_id) for sp_content in specific_contents]
        }
        # pprint(format_body_queue)
        return self._post(url, body=format_body_queue)

    def edit_queue(self, name=None, description=None):
        """Edits the queue with a new name and a new 
        descriptions

        @name: the new name of the queue
        @description: the new description of the queue

        """
        if not name or not description:
            raise OrchestratorMissingParam(value="name/description", message="name and/or description cannot be null")
        endpoint = f"/QueueDefinitions({self.id})"
        url = f"{self.base_url}{endpoint}"
        format_body_queue = {
            "Name": name,
            "Description": description
        }
        return self._put(url, body=format_body_queue)

    def delete_queue(self):
        """Deletes the queue"""
        endpoint = f"/QueueDefinitions({self.id})"
        url = f"{self.base_url}{endpoint}"
        return self._delete(url)

    def get_queue_item_comments(self, q=None):
        endpoint = "/QueueItemComments"
        if q:
            query_params = urlencode({
                "$filter": q
            })
            url = f"{self.base_url}{endpoint}?{query_params}"
        url = f"{self.base_url}{endpoint}"
        return self._get(url)



from orchestrator.orchestrator_http import OrchestratorHTTP
from orchestrator.exceptions import OrchestratorMissingParam
from urllib.parse import urlencode
import requests

__all__ = ["Process"]


class Process(OrchestratorHTTP):
    def __init__(self, client_id, refresh_token, tenant_name, folder_id=None, session=None, process_id=None, process_title=None, process_version=None, process_key=None, access_token=None):
        super().__init__(client_id=client_id, refresh_token=refresh_token, tenant_name=tenant_name, folder_id=folder_id)
        if not tenant_name or not folder_id:
            raise OrchestratorMissingParam(value="tenant_name, folder_id",
                                           message="Required parameter(s) missing: tenant_name, folder_id")
        self.tenant_name = tenant_name
        self.base_url = f"{self.cloud_url}/{self.tenant_name}/JTBOT/odata"
        self.id = process_id
        self.access_token = access_token
        self.tenant_name = tenant_name
        self.title = process_title
        self.version = process_version
        self.key = process_key
        if session:
            self.session = session
        else:
            self.session = requests.Session()

    def __str__(self):
        return f"Process Id: {self.id} \nTitle: {self.title} \nVersion: {self.version} \nKey: {self.key} \nFolder Id: {self.folder_id}"

    def info(self):
        endpoint = "/Processes"
        query_param = urlencode({
            "$filter": f"Key eq '{self.key}'"
        })
        url = f"{self.base_url}{endpoint}?{query_param}"
        return self._get(url)["value"]

    def versions(self):
        endpoint = "/Processes"
        uipath_svc = f"/UiPath.Server.Configuration.OData.GetProcessVersions(processId='{self.id}')"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        return self._get(url)["value"]

    def parameters(self):
        endpoint = "/Processes"
        uipath_svc = f"/UiPath.Server.Configuration.OData.GetArguments(key='{self.key}')"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        return self._get(url)

    def get_available_versions(self):
        """
        Gets all the available versions for a given process
        """
        endpoint = "/Processes"
        uipath_svc = f"/UiPath.Server.Configuration.OData.GetProcessVersions(processId='{self.id}')"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        return self._get(url)



from orchestrator.orchestrator_http import OrchestratorHTTP
from orchestrator.orchestrator_asset import Asset
from orchestrator.orchestrator_queue import Queue
from orchestrator.orchestrator_job import Job
from orchestrator.orchestrator_process_schedule import ProcessSchedule
from orchestrator.exceptions import OrchestratorMissingParam
from urllib.parse import urlencode
import requests


"""
Class to deal with calls and requests from a given Folder of an Orchestrator Cloud instance

"""


class Folder(OrchestratorHTTP):
    """Constructor 

    :param client_id - the client id of your organization 
    :type client_id - str 

    :param refresh_token - a refresh token. 
    :type refresh_token - str 

    :param tenant_name - your account's logical name
    :type tenant_name - str 

    :param session - an optional session object 
    :type session - Session

    :param folder_id - the id of the folder (UiPath's organizations unit)
    :type folder_id - str 

    :param folder_name - the name of the folder 
    :type folder_name : str
    """

    def __init__(self, client_id, refresh_token, tenant_name, session=None, folder_name=None,  folder_id=None, access_token=None):
        super().__init__(client_id=client_id, refresh_token=refresh_token, tenant_name=tenant_name, folder_id=folder_id, session=session)
        if not tenant_name or not folder_id:
            raise OrchestratorMissingParam(value="tenant_name",
                                           message="Required parameter missing: tenant_name")
        self.id = folder_id
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.tenant_name = tenant_name
        self.name = folder_name
        self.base_url = f"{self.cloud_url}/{self.tenant_name}/JTBOT/odata"
        if session:
            self.session = session
        else:
            self.session = requests.Session()

    def __str__(self):
        return f"Folder Id: {self.id} \nFolder Name: {self.name}"

    def info(self):
        """Returns a information of a single
            folder based on its folder id
        """
        endpoint = f"/Folders({self.id})"
        url = f"{self.base_url}{endpoint}"
        data = self._get(url)
        return data

    def get_queues(self, options=None):
        """Parameters:
            :param options - dictionary of filtering odata options
            :type options - dict
        """
        endpoint = "/QueueDefinitions"
        if options:
            query_params = urlencode(options)
            url = f"{self.base_url}{endpoint}?{query_params}"
        else:
            url = f"{self.base_url}{endpoint}"
        data = self._get(url)
        filt_data = data['value']
        return [Queue(self.client_id, self.access_token, self.tenant_name, self.id, self.name, self.session, queue["Name"],  queue["Id"], access_token=self.access_token) for queue in filt_data]

    def get_queue_ids(self, options=None):
        """
            Returns a list of dictionaries containing
            the queue name and the queue id

            :options dictionary for odata options

        """
        queues = self.get_queues(options)
        ids = {}
        for queue in queues:
            ids.update({queue.id: queue.name})
        return ids

    def get_processing_records(self, options=None):
        """Returns a list of queue processing records for all the queues

            :param options - dictionary for odata options
            :type options - dict


        """
        endpoint = "/QueueProcessingRecords"
        uipath_svc = "/UiPathODataSvc.RetrieveQueuesProcessingStatus"
        if options:
            query_params = urlencode(options)
            url = f"{self.base_url}{endpoint}?{query_params}"
        else:
            url = f"{self.base_url}{endpoint}{uipath_svc}"
        data = self._get(url)
        return data['value']

    def get_queue_by_id(self, queue_id):
        queues = self.get_queue_ids()
        return Queue(self.client_id, self.refresh_token, self.tenant_name, self.id, self.name, self.session, queues[queue_id], queue_id=int(queue_id), access_token=self.access_token)

    def get_assets(self, options=None):
        """
            Returns list of assets
            :options dict of odata filter options
        """
        endpoint = "/Assets"
        if options:
            query_params = urlencode(options)
            url = f"{self.base_url}{endpoint}?{query_params}"
        else:
            url = f"{self.base_url}{endpoint}"
        data = self._get(url)
        # pprint(data)
        # pprint(self.id)
        filt_data = data['value']
        return [Asset(self.client_id, self.refresh_token, self.tenant_name, self.id, self.name, self.session, asset["Id"], asset["Name"], access_token=self.access_token) for asset in filt_data]

    def get_asset_ids(self, options=None):
        """
            Returns a dictionary of ky value pairs
            where keys are the assets names and the values
            are the ids


            :param options - dictionary of odata filter options
        """

        assets = self.get_assets(options)
        ids = {}
        for asset in assets:
            ids.update({asset.id: asset.name})
        return ids

    def get_asset_by_id(self, asset_id):
        assets = self.get_asset_ids()
        return Asset(self.client_id, self.refresh_token, self.tenant_name, self.id, self.name, self.session, asset_id, assets[asset_id], access_token=self.access_token)

    def create_asset(self, body=None):
        pass

    def get_process_schedules(self, options=None):
        endpoint = "/ProcessSchedules"
        if options:
            query_params = urlencode(options)
            url = f"{self.base_url}{endpoint}?{query_params}"
        else:
            url = f"{self.base_url}{endpoint}"
        data = self._get(url)
        filt_data = data["value"]
        return [ProcessSchedule(self.client_id, self.refresh_token, self.tenant_name, self.folder_id, self.session, process["Id"], process["Name"], access_token=self.access_token) for process in filt_data]

    def get_schedule_ids(self, options=None):
        """
            Returns a list of dictionaries
                name -- schedule_id
        """
        process_schedules = self.get_process_schedules(options=options)
        ids = {}
        for schedule in process_schedules:
            ids.update({schedule.id: schedule.name})
        return ids

    def get_sessions(self, options=None):
        """
            Gets all the sessions for the current folder
        """
        endpoint = "/Sessions"
        if options:
            query_params = urlencode(options)
            url = f"{self.base_url}{endpoint}?{query_params}"
        else:
            url = f"{self.base_url}{endpoint}"
        return self._get(url)["value"]

    def get_machine_runtime_sessions(self):
        """
            No se por que no va
        """
        endpoint = "/Sessions"
        uipath_svc = "UiPath.Server.Configuration.OData.GetMachineSessionRuntimesByFolderId(folderId={self.id})"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        return self._get(url)

    def get_jobs(self, top="100", options=None):
        """
        Returns the jobs of a given folder

        @top : maximum number of results (100 default)
        @options: dictionary of odata filtering options
        """
        endpoint = "/Jobs"
        default = {"$orderby": "StartTime desc", "$top": f"{top}"}
        enc_default = urlencode(default)
        if options:
            default.update(options)
            query_params = urlencode(default)
            url = f"{self.base_url}{endpoint}?{query_params}"
        else:
            url = f"{self.base_url}{endpoint}?{enc_default}"
        data = self._get(url)["value"]
        # print(len(data))
        return [Job(self.client_id, self.refresh_token, self.tenant_name, self.id, self.name, self.session, job["Id"], job["Key"], job["ReleaseName"], access_token=self.access_token) for job in data]

    def get_job_keys(self, top="100", options=None):
        """

        """
        jobs = self.get_jobs(top, options)
        ids = {}
        for job in jobs:
            ids.update({job.key: job.name})
        return ids

    def get_job_by_key(self, key):
        endpoint = "/Jobs"
        query_param = urlencode({"$filter": f"Key eq {key}"})
        url = f"{self.base_url}{endpoint}?{query_param}"
        data = self._get(url)["value"][0]
        return Job(self.client_id, self.refresh_token, self.tenant_name, self.id, self.name, self.session, data["Id"], data["Key"], data["ReleaseName"], access_token=self.access_token)

    def job_triggers(self, options=None):
        endpoint = "/JobTriggers"
        if options:
            query_params = urlencode(options)
            url = f"{self.base_url}{endpoint}?{query_params}"
        else:
            url = f"{self.base_url}{endpoint}"
        return self._get(url)["value"]



from orchestrator.orchestrator_http import OrchestratorHTTP
from orchestrator.exceptions import OrchestratorMissingParam
import requests

__all__ = ["ProcessSchedule"]


class ProcessSchedule(OrchestratorHTTP):
    def __init__(self, client_id, refresh_token, tenant_name, folder_id=None, session=None, process_id=None, process_name=None, access_token=None):
        super().__init__(client_id=client_id, refresh_token=refresh_token, tenant_name=tenant_name, folder_id=folder_id)
        if not tenant_name or not folder_id:
            raise OrchestratorMissingParam(value="tenant_name, folder_id",
                                           message="Required parameter(s) missing: tenant_name, folder_id")
        self.tenant_name = tenant_name
        self.access_token = access_token
        self.id = process_id
        self.name = process_name
        self.folder_id = folder_id
        self.base_url = f"{self.cloud_url}/{self.tenant_name}/JTBOT/odata"
        if session:
            self.session = session
        else:
            self.session = requests.Session()

    def __str__(self):

        return f"Schedule Id: {self.id} \nName: {self.name}"

    def info(self):
        endpoint = f"/ProcessSchedules({self.id})"
        url = f"{self.base_url}{endpoint}"
        return self._get(url)

    def schedule(self):
        info = self.info()
        return info["StartProcessCronSummary"]

    def delete(self):
        """
            Deletes a schedule
        """
        endpoint = f"/ProcessSchedules({self.id})"
        url = f"{self.base_url}{endpoint}"
        return self._delete(url)

    def enable(self, enable: bool = False, schedule_ids=[]):
        schedule_ids.append(self.id)
        endpoint = "/ProcessSchedules"
        uipath_svc = "/UiPath.Server.Configuration.OData.SetEnabled"
        body = {
            "enabled": enable,
            "scheduleIds": schedule_ids
        }
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        return self._post(url, body=body)



from orchestrator.orchestrator_http import OrchestratorHTTP
import requests
from orchestrator.exceptions import OrchestratorMissingParam

"""
    Asset class for dealing with Orchestrator requests which involve a given asset
"""


class Asset(OrchestratorHTTP):
    """
    Container class for asset entities in Orchestrator Cloud. 

    :param client_id - the client id of your cloud organization unit.
    :type client_id - str

    :param refresh_token - a refresh token. 
    :type refresh_token - str 

    :param tenant_name - your account's logical name
    :type tenant_name - str 

    :param folder_id - the id of the folder (UiPath's organizations unit)
    :type folder_id - str 

    :param folder_name - the name of the folder 
    :type folder_name : str 

    :param session - a session object to deal with the requests 
    :type session - Session 

    :param asset_id - the id of the asset 
    :type asset_id - int

    :param asset_name - the name of the asset
    :type asset_id - str 

    """

    def __init__(self, client_id, refresh_token, tenant_name, folder_id=None, folder_name=None, session=None, asset_id=None, asset_name=None, access_token=None):
        """Constructor"""
        super().__init__(client_id=client_id, refresh_token=refresh_token, tenant_name=tenant_name, folder_id=folder_id, session=session)
        if not asset_id:
            raise OrchestratorMissingParam(value="asset_id",
                                           message="Required parameter(s) missing: asset_id")
        self.tenant_name = tenant_name
        self.base_url = f"{self.cloud_url}/{self.tenant_name}/JTBOT/odata"
        self.folder_id = folder_id
        self.access_token = access_token
        self.folder_name = folder_name
        self.id = asset_id
        self.name = asset_name
        if session:
            self.session = session
        else:
            self.session = requests.Session()

    def __str__(self):
        return f"Asset Id: {self.id} \nAsset Name: {self.name} \nFolder Id: {self.folder_id} \nFolder Name: {self.folder_name}"

    def info(self):
        """
        Gives back information about the current asset.
        """
        endpoint = f"/Assets({self.id})"
        url = f"{self.base_url}{endpoint}"
        return self._get(url)

    def edit(self, body=None):
        """
            Edits an asset 

            :param body - the body of the request with the attributes changed
            :type body - dict
        """
        endpoint = f"/Assets({self.id})"
        url = f"{self.base_url}{endpoint}"
        return self._put(url, body=body)

    def delete(self, body=None):
        endpoint = f"/Assets({self.id})"
        url = f"{self.base_url}{endpoint}"
        return self._delete(url, body=body)



from orchestrator.orchestrator_http import OrchestratorHTTP
import requests
from orchestrator.exceptions import OrchestratorMissingParam


class Library(OrchestratorHTTP):
    def __init__(self, client_id, refresh_token, tenant_name, session=None, lib_key=None, lib_id=None, lib_title=None, folder_id=None, access_token=None):
        super().__init__(client_id=client_id, refresh_token=refresh_token, tenant_name=tenant_name, folder_id=folder_id, session=session)
        if not lib_key:
            raise OrchestratorMissingParam(value="library key",
                                           message="Required parameter(s) missing: library key")
        self.tenant_name = tenant_name
        self.access_token = access_token
        self.base_url = f"{self.cloud_url}/{self.tenant_name}/JTBOT/odata"
        self.folder_id = folder_id
        self.id = lib_id
        self.key = lib_key
        self.name = lib_title
        if session:
            self.session = session
        else:
            self.session = requests.Session()

    def __str__(self):
        idx = f"Id: {self.id}\n"
        key = f"Key: {self.key}\n"
        title = f"Title: {self.name}\n"

        return f"{idx}{key}{title}"



from platform import machine
from orchestrator.exceptions import OrchestratorMissingParam
from orchestrator.orchestrator_http import OrchestratorHTTP
import requests
from urllib.parse import urlencode


class Machine(OrchestratorHTTP):
    """
    Constructor. 

    @client_id: the client id 
    @refresh_token: a refresh token  
    @tenant_name: account's logical name
    @folder_id: the folder id 
    @folder_name: the folder name
    @session: a session object (options)
    @queue_name: the queue name
    @queue_id: the queue id
    """

    def __init__(self, client_id, refresh_token, tenant_name, folder_id=None, session=None, machine_id=None, machine_key=None, machine_name=None, access_token=None):
        super().__init__(client_id=client_id, refresh_token=refresh_token, tenant_name=tenant_name, folder_id=folder_id, session=session)
        if not machine_id:
            raise OrchestratorMissingParam(value="queue_id",
                                           message="Required parameter(s) missing: queue_id")
        self.id = machine_id
        self.access_token = access_token
        self.name = machine_name
        self.key = machine_key
        self.folder_id = folder_id
        self.tenant_name = tenant_name
        self.base_url = f"{self.cloud_url}/{self.tenant_name}/JTBOT/odata"
        if session:
            self.session = session
        else:
            self.session = requests.Session()

    def info(self):
        endpoint = f"/Machines({self.id})"
        url = f"{self.base_url}{endpoint}"
        return self._get(url)



from pprint import pprint
from orchestrator.orchestrator_http import OrchestratorHTTP
from orchestrator.exceptions import OrchestratorMissingParam
import requests
from urllib.parse import urlencode
import json

__all__ = ["QueueItem"]


class QueueItem(OrchestratorHTTP):

    def __init__(self, client_id, refresh_token, tenant_name, folder_id=None, folder_name=None, queue_name=None, queue_id=None, session=None, item_id=None, content=None, reference=None, access_token=None):
        super().__init__(client_id=client_id, refresh_token=refresh_token, tenant_name=tenant_name, folder_id=folder_id,
                         session=session)
        if not item_id:
            raise OrchestratorMissingParam(value="item id",
                                           message="Required parameter(s) missing: item_id")
        self.specific_content = content
        self.access_token = access_token
        self.reference = reference
        self.tenant_name = tenant_name
        self.folder_id = folder_id
        self.folder_name = folder_name
        self.queue_name = queue_name
        self.queue_id = queue_id
        self.id = item_id

        self.base_url = f"{self.cloud_url}/{self.tenant_name}/JTBOT/odata"
        if session:
            self.session = session
        else:
            self.session = requests.Session()

    def __str__(self):
        return f"Item Id: {self.id} \nQueue: {self.queue_name} \nFolder: {self.folder_name}"

    def content(self):
        info = self.info()
        return info["SpecificContent"]

    def info(self):
        """
            Gets a single item by item id

            Parameters:

            :param item_id : item id


            Necesito una clase Item
        """
        endpoint = f"/QueueItems({self.id})"
        url = f"{self.base_url}{endpoint}"
        return self._get(url)

    def delete(self):
        """
            Deletes the given queue item
        """
        endpoint = f"/QueueItems({self.id})"
        url = f"{self.base_url}{endpoint}"
        return self._delete(url)

    def edit(self, body=None):
        """
            edits the given queue item
        """
        endpoint = f"/QueueItems({self.id})"
        url = f"{self.base_url}{endpoint}"
        return self._put(url, body=body)

    def last_entry(self):
        """
            Returns the last entry of the given
            queue item
        """
        endpoint = f"/QueueItems({self.id})"
        uipath_svc = "/UiPath.Server.Configuration.OData.GetItemLastRetry"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        return self._get(url)

    def history(self):
        """
            Returns the history of the given queue
            item
        """
        endpoint = f"/QueueItems({self.id})"
        uipath_svc = "/UiPathODataSvc.GetItemProcessingHistory"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        return self._get(url)["value"]

    def set_transaction_progress(self, status=None):
        """
            Updates the progress field of a given queue
            item (note: it must be already In Progress)
        """
        if not status:
            raise OrchestratorMissingParam(value="status", message="status cannot be None")
        endpoint = f"/QueueItems({self.id})"
        uipath_svc = "/UiPathODataSvc.SetTransactionProgress"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        body = {
            "progress": status
        }
        return self._post(url, body=body)

    def set_transaction_status(self, success: bool, reason=None, details=None, exception_type=None, fail_reason=None):
        endpoint = f"/Queues({self.id})"
        uipath_svc = "/UiPathODataSvc.SetTransactionResult"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        if success:
            transaction_body = {
                "transactionResult": {
                    "IsSuccessful": True,
                    # "ProcessingException": {
                    #     "Reason": reason,
                    #     "Details": details,
                    #     "Type": exception_type,

                    # },
                    # "Output": {
                    #     "fail_reason": fail_reason
                    # }

                }
            }
        else:
            transaction_body = {
                "transactionResult": {
                    "IsSuccessful": False,
                    "ProcessingException": {
                        "Reason": reason,
                        "Details": details,
                        "Type": exception_type,

                    },
                    "Output": {
                        "fail_reason": fail_reason
                    }

                }
            }
        # pprint(transaction_body)
        return self._post(url, body=transaction_body)

    def events(self):
        """
            Gets queue item events associated to the current
            queue item

            No funciona no se por que
        """
        endpoint = "/QueueItemEvents"
        uipath_svc = f"/UiPathODataSvc.GetQueueItemEventsHistory(queueItemId={{self.id}})"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        return self._get(url)

    def make_comment(self, text=None):
        body = {
            "Text": text,
            "QueueItemId": self.id
        }
        endpoint = "/QueueItemComments"
        url = f"{self.base_url}{endpoint}"
        return self._post(url, body=body)




from typing import List
from orchestrator.orchestrator_http import OrchestratorHTTP
import requests
import json
from urllib.parse import urlencode
from orchestrator.orchestrator_folder import Folder
from orchestrator.orchestrator_library import Library
from orchestrator.orchestrator_process import Process
from orchestrator.orchestrator_machine import Machine


__all__ = ["Orchestrator"]

"""
Parent class for initializing a client for UiPath's Orchestrator's API
"""


class Orchestrator(OrchestratorHTTP):
    """
    Constructor. 

    @client_id: the client id 
    @refresh_token: a refresh token  
    @tenant_name: account's logical name
    @folder_id: a folder id (optional)
    @session: a session object (options)
    @file: a credentials file containing client_id, refresh_token and tenant_name (optional) 
    """

    def __init__(
        self,
        client_id=None,
        refresh_token=None,
        tenant_name=None,
        folder_id=None,
        session=None,
        file=None

    ):

        super().__init__(client_id=client_id, refresh_token=refresh_token, tenant_name=tenant_name, folder_id=folder_id, session=session, file=file)
        # if not client_id or not refresh_token:
        #     raise OrchestratorAuthException(
        #         value=None, message="client id and refresh token cannot be left empty"
        #     )
        # else:
        #     self.client_id = client_id
        if file:
            self.base_url = f"{self.cloud_url}/{self.tenant_name}/JTBOT/odata"

            try:
                f = open(file)
                data = json.load(f)
                self.client_id = data["client_id"]
                self.refresh_token = data["refresh_token"]
                self.tenant_name = data["tenant_name"]
            except KeyError:
                raise
        else:
            self.client_id = client_id
            self.refresh_token = refresh_token
            self.tenant_name = tenant_name
            self.folder_id = folder_id
            self.base_url = f"{self.cloud_url}/{self.tenant_name}/JTBOT/odata"
        if session:
            # print("session set")
            self.session = session
        else:
            # print("session not set")
            self.session = requests.Session()

    def __str__(self):
        if self.folder_id:
            return f"Folder Id: {self.folder_id} \nTenant: {self.tenant_name}"
        return {f"Tenant: {self.tenant_name}"}

    def get_folders(self, options=None):
        """
        Gets all the folders from a given organization

        @options: dictionary of odata filtering options
        ========
        @returns: a list of Folders of the given organization
        """
        endpoint = "/Folders"
        if options:
            query_params = urlencode(options)
            url = f"{self.base_url}{endpoint}?{query_params}"
        else:
            url = f"{self.base_url}{endpoint}"
        data = self._get(url)
        filt_data = data['value']
        return [Folder(self.client_id, self.refresh_token, self.tenant_name, self.session, folder["DisplayName"], folder["Id"], access_token=self.access_token) for folder in filt_data]

    def get_folder_ids(self, options=None):
        """
            Returns a dictionary with the folder ids 

            @options: dictionary of odata filtering options
            =======
            @returns: a dictionary where the keys are the ids 
            and the values the names of the folders in the given
            organization 
        """
        folders = self.get_folders(options)
        ids = {}
        for folder in folders:
            ids.update({folder.id: folder.name})
        return ids

    def get_folder_by_id(self, folder_id):
        """
        Returns a single folder by its id 

        @folder_id: the id of the folder
        ==========
        @returns: a Folder object with the specified folder id
        """
        ids = self.get_folder_ids()
        self.folder_id = folder_id
        folder_name = ids[folder_id]
        return Folder(client_id=self.client_id, refresh_token=self.refresh_token, tenant_name=self.tenant_name,  session=self.session, folder_name=folder_name, folder_id=int(folder_id), access_token=self.access_token)

    def get_folder_by_name(self, folder_name):
        """
            Returns a single folder by its name

            @folder_name: the name of the folder
            ============
            @returns: a Folder object with the specified folder_name
        """
        pass

    def usernames(self, options=None):
        """
            No se por que no va
        """
        endpoint = "/Sessions"
        uipath_svc = "/UiPath.Server,Configuration.OData.GetUsernames"
        if options:
            query_params = urlencode(options)
            url = f"{self.base_url}{endpoint}{uipath_svc}?{query_params}"
        else:
            url = f"{self.base_url}{endpoint}{uipath_svc}"
        return self._get(url)

    def permissions(self, options=None):
        endpoint = "/Permissions"
        if options:
            query_params = urlencode(options)
            url = f"{self.base_url}{endpoint}?{query_params}"
        else:
            url = f"{self.base_url}{endpoint}"
        return self._get(url)

    def get_processes(self, options=None):
        """
        Gets all the processes of a given organization

        @options: a dictionary of odata filtering options
        ========
        @returns: a list of Processes of the given organization
        """
        endpoint = "/Processes"
        if options:
            query_params = urlencode(options)
            url = f"{self.base_url}{endpoint}?{query_params}"
        else:
            url = f"{self.base_url}{endpoint}"
        processes = self._get(url)["value"]
        return [Process(self.client_id, self.refresh_token, self.tenant_name, self.folder_id, self.session, process["Id"], process["Title"], process["Version"], process["Key"], access_token=self.access_token) for process in processes]

    def get_processes_keys(self, options=None):
        """
            Returns a dictionary with the processes keys

            @options: dictionary of odata filtering options
            ========
            @returns: a dictionary where the keys are the process'
            key and the values the process' title of the processes 
            in the given organization
        """
        processes = self.get_processes(options=options)
        ids = {}
        for process in processes:
            ids.update({process.key: process.title})
        return ids

    def get_process_by_key(self, process_key):
        """
        Returns a single process by is key

        @process_key: the key of the process 
        ============
        @returns: a Process object with the specified process key 
        """
        query_param = urlencode({
            "$filter": f"Key eq '{process_key}"
        })
        endpoint = "/Processes"
        url = f"{self.base_url}{endpoint}?{query_param}"
        process = self._get(url)["value"][0]
        return Process(self.client_id, self.refresh_token, self.tenant_name, self.folder_id,
                       self.session, process["Id"], process["Title"], process["Version"], process["Key"], access_token=self.access_token)

    def get_libraries(self, options=None):
        """
        Gets all the libraries of a given organization

        @options: a dictionary of odata filtering options
        ========
        @returns: a list of Libraries of the given organization
        """
        endpoint = "/Libraries"
        if options:
            query_params = urlencode(options)
            url = f"{self.base_url}{endpoint}?{query_params}"
        else:
            url = f"{self.base_url}{endpoint}"
        libraries = self._get(url)["value"]
        return [Library(self.client_id, self.refresh_token, self.tenant_name, self.session,  lib["Key"], lib["Id"], lib["Title"], self.folder_id, access_token=self.access_token) for lib in libraries]

    def get_machines(self, options=None):
        """
        Gets all the machines of a given organization 
        """
        endpoint = "/Machines"
        if options:
            query_params = urlencode(options)
            url = f"{self.base_url}{endpoint}?{query_params}"
        else:
            url = f"{self.base_url}{endpoint}"
        data = self._get(url)["value"]
        return [Machine(self.client_id, self.refresh_token, self.tenant_name, self.folder_id, self.session, machine["Id"], machine["Key"], machine["Name"]) for machine in data]

    def get_machine_ids(self, options=None):
        """
        Returns a dictionary of the machine keys and their
        names 

        @options: dictionary of odata filtering options 
        """
        machines = self.get_machines(options)
        ids = {}
        for machine in machines:
            ids.update({machine.id: machine.name})
        return ids

    def get_machine_by_id(self, machine_id):
        """
        Returns a single machine by its id 

        @machine_id: the id of the machine 
        """

        endpoint = f"/Machines({machine_id})"
        url = f"{self.base_url}{endpoint}"
        machine = self._get(url)
        return Machine(self.client_id, self.refresh_token, self.tenant_name, self.folder_id, self.session, machine_id, machine["Key"], machine["Name"], access_token=self.access_token)



from pprint import pprint
from orchestrator.orchestrator_http import OrchestratorHTTP
import requests
from orchestrator.exceptions import OrchestratorMissingParam
from urllib.parse import urlencode
from orchestrator.orchestrator_logs import Log


class Job(OrchestratorHTTP):
    def __init__(self, client_id, refresh_token, tenant_name, folder_id=None, folder_name=None, session=None, job_id=None, job_key=None, job_name=None, access_token=None):
        super().__init__(client_id=client_id, refresh_token=refresh_token, tenant_name=tenant_name, folder_id=folder_id, session=session)
        if not job_key:
            raise OrchestratorMissingParam(value="asset_id",
                                           message="Required parameter(s) missing: asset_id")
        self.tenant_name = tenant_name
        self.access_token = access_token
        self.base_url = f"{self.cloud_url}/{self.tenant_name}/JTBOT/odata"
        self.folder_id = folder_id
        self.folder_name = folder_name
        self.id = job_id
        self.key = job_key
        self.name = job_name
        if session:
            self.session = session
        else:
            self.session = requests.Session()

    def __str__(self):
        return f"Key: {self.key}\nName: {self.name}"

    def info(self):
        endpoint = f"/Jobs({self.id})"
        url = f"{self.base_url}{endpoint}"
        return self._get(url)

    def stop(self):
        """
        Stops the given job
        """
        endpoint = f"/Jobs({self.id})"
        uipath_svc = "/UiPath.Server.Configuration.OData.StopJob"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        cancel_body = {
            "strategy": "SoftStop"
        }
        return self._post(url, body=cancel_body)

    def kill(self):
        """
        Kill the given job
        """
        endpoint = f"/Jobs({self.id})"
        uipath_svc = "/UiPath.Server.Configuration.OData.StopJob"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        cancel_body = {
            "strategy": "Kill"
        }
        return self._post(url, body=cancel_body)

    def restart(self):
        """
        Restarts the given job
        """
        endpoint = f"/Jobs"
        uipath_svc = "/UiPath.Server.Configuration.OData.RestartJob"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        restart_body = {
            "jobId": self.id
        }
        return self._post(url, body=restart_body)

    def resume(self):
        """
        Restarts the given job
        """
        endpoint = f"/Jobs"
        uipath_svc = "/UiPath.Server.Configuration.OData.ResumeJob"
        url = f"{self.base_url}{endpoint}{uipath_svc}"
        resume_body = {
            "jobKey": self.key
        }
        return self._post(url, body=resume_body)

    def get_logs(self, trace="Info"):
        # need a Log Message class
        endpoint = "/RobotLogs"
        query_param = urlencode({
            "$filter": f"ProcessName eq '{self.name}' and Level eq '{trace}' and JobKey eq {self.key}",
            "$orderby": "TimeStamp desc"
        })
        url = f"{self.base_url}{endpoint}?{query_param}"
        logs = self._get(url)["value"]
        # pprint(data[0])
        return [Log(self.client_id, self.refresh_token, self.tenant_name, self.folder_id, self.folder_name, self.session, log["Message"], log["Level"], self.key, log["TimeStamp"]) for log in logs]



from datetime import datetime
import requests
import random
import json
import string
from pprint import pprint
import logging

from orchestrator.exceptions import OrchestratorAuthException, OrchestratorMissingParam


class OrchestratorHTTP(object):
    cloud_url = "https://cloud.uipath.com"
    account_url = "https://account.uipath.com"
    oauth_endpoint = "/oauth/token"
    _now = datetime.now()
    _token_expires = datetime.now()
    access_token = None

    def __init__(
        self,
        client_id=None,
        refresh_token=None,
        tenant_name=None,
        folder_id=None,
        session=None,
        file=None

    ):
        if not client_id or not refresh_token:
            if file:
                f = open(file)
                try:
                    data = json.load(f)
                    self.client_id = data["client_id"]
                    self.refresh_token = data["refresh_token"]
                    self.tenant_name = data["tenant_name"]
                    self.base_url = f"{self.cloud_url}/{self.tenant_name}/JTBOT/odata"
                    self.folder_id = data["folder_id"]
                except KeyError as err:
                    print(err)
                    raise
            else:
                raise OrchestratorAuthException(
                    value=None, message="client id and refresh token cannot be left empty"
                )

        else:
            self.client_id = client_id
            self.refresh_token = refresh_token
            self.folder_id = folder_id
            self.tenant_name = tenant_name
            self.base_url = f"{self.cloud_url}/{self.tenant_name}/JTBOT/odata"
        if session:
            self.session = session
        else:
            self.session = requests.Session()
        self.expired = True

    @staticmethod
    def generate_reference():
        return ''.join(random.choices(string.ascii_uppercase + string.digits, k=10))

    def _get_token(self):
        body = {
            "grant_type": "refresh_token",
            "client_id": self.client_id,
            "refresh_token": self.refresh_token,
        }
        headers = {"Content-Type": "application/json"}
        url = f"{self.account_url}{self.oauth_endpoint}"
        try:
            r = self.session.post(url=url, data=json.dumps(body), headers=headers)
            token_data = r.json()
            token = token_data["access_token"]
            expiracy = token_data["expires_in"]
            self.access_token = token
            self._token_expires = expiracy
        except Exception as err:
            print(err)

    def _auth_header(self):
        return {"Authorization": f"Bearer {self.access_token}"}

    @staticmethod
    def _content_header():
        return {"Content-Type": "application/json"}

    def _folder_header(self):
        if not self.folder_id:
            raise OrchestratorAuthException(value="folder id", message="folder cannot be null")
        return {"X-UIPATH-OrganizationUnitId": f"{self.folder_id}"}

    def _internal_call(self, method, endpoint, *args, **kwargs):
        # pprint(self.folder_id)
        # if self.expired:
        #     logging.debug("Token has expired")
        #     self._get_token()
        #     self.expired = False

        headers = self._auth_header()
        # pprint(headers)
        if method == "POST":
            headers.update(self._content_header())
        if self.folder_id:
            headers.update(self._folder_header())
        try:
            # print(endpoint)
            if kwargs:
                # pprint(kwargs)
                item_data = kwargs['body']['body']
                # pprint(json.dumps(item_data))
                # pprint(item_data)
                r = self.session.request(method, endpoint, json=item_data, headers=headers)
            else:
                r = self.session.request(method, endpoint, headers=headers)
                # print(r.status_code)
                if r.status_code == 401:
                    self._get_token()
                    headers = self._auth_header()
                    r_retry = self.session.request(method, endpoint, headers=headers)
                    return r_retry.json()
                if r.status_code not in range(200, 400):
                    logging.error(f"An error ocurred.\nStatus code: {r.status_code}")
                    # print(r.json())
            # print(endpoint)
            logging.debug(f"{r.status_code} ---- {r.url}")
            try:
                # pprint(r.json())
                return r.json()
            except requests.exceptions.JSONDecodeError:
                return r.text
        except Exception as err:
            print(err)
            raise

    def _get(self, url, *args, **kwargs):

        return self._internal_call("GET", url, args, kwargs)

    def _post(self, url, *args, **kwargs):
        # pprint(kwargs)
        return self._internal_call("POST", url, args, body=kwargs)

    def _put(self, url, *args, **kwargs):
        return self._internal_call("PUT", url, args, body=kwargs)

    def _delete(self, url, *args, **kwargs):
        return self._internal_call("DELETE", url, args, kwargs)



from datetime import datetime
from orchestrator.orchestrator_http import OrchestratorHTTP
import requests
from orchestrator.exceptions import OrchestratorMissingParam


class Log(OrchestratorHTTP):
    def __init__(self, client_id, refresh_token, tenant_name, folder_id=None, folder_name=None, session=None, msg=None, trace=None, key=None, stamp=None):
        super().__init__(client_id=client_id, refresh_token=refresh_token, tenant_name=tenant_name, folder_id=folder_id, session=session)
        if not key:
            raise OrchestratorMissingParam(value="job key",
                                           message="Required parameter(s) missing: key")
        self.tenant_name = tenant_name
        self.base_url = f"{self.cloud_url}/{self.tenant_name}/JTBOT/odata"
        self.folder_id = folder_id
        self.folder_name = folder_name
        self.key = key
        self.message = msg
        self.timestamp = stamp
        self.trace = trace
        if session:
            self.session = session
        else:
            self.session = requests.Session()

    def __str__(self):
        return f"Job Key: {self.key}\nMessage: {self.message}\nTimeStamp: {self.timestamp}\nTrace: {self.trace}"

    def create(self, level="Info", message=None):
        endpoint = "/api/Logs/SubmitLogs"
        body = {
            "message": message,
            "level": level,
            "timeStamp": datetime.now,
            "jobId": self.id
        }
        url = f"{self.base_url}{endpoint}"
        return self._post(url, body=body)



class OrchestratorAuthException(Exception):
    def __init__(self, value, message):
        super().__init__(self, message)
        self.value = value

class OrchestratorFormatException(Exception):
    def __init__(self, value, message):
        super().__init__(self, message)
        self.value = value


class OrchestratorMissingParam(Exception):
    def __init__(self, value, message):
        super().__init__(self, message)
        self.value = value
